@@ -379,6 +379,61 @@ class _UuidUuidKeysMixin(object):
         return uuid.UUID(bytes=data1), uuid.UUID(bytes=data2)
 
 
+class _UuidUuidUuidKeysMixin(object):
+    def _serialize_key(self, key1_key2_key3):
+        assert type(key1_key2_key3) == tuple and len(key1_key2_key3) == 3
+        key1, key2, key3 = key1_key2_key3
+
+        if key1 is None:
+            key1 = uuid.UUID(bytes=b'\x00' * 16)
+        if key2 is None:
+            key2 = uuid.UUID(bytes=b'\x00' * 16)
+        if key3 is None:
+            key3 = uuid.UUID(bytes=b'\x00' * 16)
+
+        assert isinstance(key1, uuid.UUID)
+        assert isinstance(key2, uuid.UUID)
+        assert isinstance(key3, uuid.UUID)
+
+        return key1.bytes + key2.bytes + key3.bytes
+
+    def _deserialize_key(self, data):
+        assert type(data) == bytes
+        assert len(data) == 48
+
+        data1, data2, data3 = data[0:16], data[16:32], data[32:48]
+        return uuid.UUID(bytes=data1), uuid.UUID(bytes=data2), uuid.UUID(bytes=data3)
+
+
+class _UuidUuidUuidUuidKeysMixin(object):
+    def _serialize_key(self, key1_key2_key3_key4):
+        assert type(key1_key2_key3_key4) == tuple and len(key1_key2_key3_key4) == 4
+        key1, key2, key3, key4 = key1_key2_key3_key4
+
+        if key1 is None:
+            key1 = uuid.UUID(bytes=b'\x00' * 16)
+        if key2 is None:
+            key2 = uuid.UUID(bytes=b'\x00' * 16)
+        if key3 is None:
+            key3 = uuid.UUID(bytes=b'\x00' * 16)
+        if key4 is None:
+            key4 = uuid.UUID(bytes=b'\x00' * 16)
+
+        assert isinstance(key1, uuid.UUID)
+        assert isinstance(key2, uuid.UUID)
+        assert isinstance(key3, uuid.UUID)
+        assert isinstance(key4, uuid.UUID)
+
+        return key1.bytes + key2.bytes + key3.bytes + key4.bytes
+
+    def _deserialize_key(self, data):
+        assert type(data) == bytes
+        assert len(data) == 64
+
+        data1, data2, data3, data4 = data[0:16], data[16:32], data[32:48], data[48:64]
+        return uuid.UUID(bytes=data1), uuid.UUID(bytes=data2), uuid.UUID(bytes=data3), uuid.UUID(bytes=data4)
+
+
 class _Uint16UuidTimestampKeysMixin(object):
     @staticmethod
     def new_key():
@@ -761,9 +816,12 @@ class _UuidStringKeysMixin(object):
 
     def _deserialize_key(self, data):
         assert type(data) == bytes
-        assert len(data) > 16
-        data1, data2 = data[:16], data[16:]
-
+        assert len(data) >= 16
+        data1 = data[:16]
+        if len(data) > 16:
+            data2 = data[16:]
+        else:
+            data2 = b''
         return uuid.UUID(bytes=data1), data2.decode('utf8')
 
 
@@ -845,6 +903,26 @@ class _Bytes32UuidKeysMixin(object):
         return data1, uuid.UUID(bytes=data2)
 
 
+class _UuidBytes32KeysMixin(object):
+    def _serialize_key(self, key1_key2):
+        assert type(key1_key2) == tuple and len(key1_key2) == 2
+        key1, key2 = key1_key2
+
+        assert isinstance(key1, uuid.UUID)
+
+        assert type(key2) == bytes
+        assert len(key2) == 32
+
+        return key2.bytes + key1
+
+    def _deserialize_key(self, data):
+        assert type(data) == bytes
+        assert len(data) == 48
+
+        data1, data2 = data[0:16], data[16:48]
+        return uuid.UUID(bytes=data1), data2
+
+
 class _Bytes32StringKeysMixin(object):
     def _serialize_key(self, key1_key2):
         assert type(key1_key2) == tuple and len(key1_key2) == 2
@@ -890,10 +968,10 @@ class _UuidUuidUuidStringKeysMixin(object):
         assert type(key1_key2_key3_key4) == tuple and len(key1_key2_key3_key4) == 4
         key1, key2, key3, key4 = key1_key2_key3_key4
 
-        assert isinstance(key1, uuid.UUID)
-        assert isinstance(key2, uuid.UUID)
-        assert isinstance(key3, uuid.UUID)
-        assert type(key4) == str
+        assert isinstance(key1, uuid.UUID), 'key1 must be a UUID, was {}: {}'.format(type(key1), key1)
+        assert isinstance(key2, uuid.UUID), 'key2 must be a UUID, was {}: {}'.format(type(key2), key2)
+        assert isinstance(key3, uuid.UUID), 'key3 must be a UUID, was {}: {}'.format(type(key3), key3)
+        assert type(key4) == str, 'key4 must be a str, was {}: {}'.format(type(key4), key4)
 
         return key1.bytes + key2.bytes + key3.bytes + key4.encode('utf8')
 
@@ -1051,6 +1129,41 @@ class _Bytes16TimestampKeysMixin(object):
         return key1, key2
 
 
+class _Bytes16TimestampUuidKeysMixin(object):
+    @staticmethod
+    def new_key():
+        return os.urandom(20), np.datetime64(time_ns(), 'ns'), uuid.uuid4()
+
+    def _serialize_key(self, keys):
+        assert type(keys) == tuple, 'keys in {}._serialize_key must be a tuple, was: "{}"'.format(
+            self.__class__.__name__, keys)
+        assert len(keys) == 3
+        key1, key2, key3 = keys
+
+        if not key1:
+            key1 = b'\x00' * 16
+
+        assert key1 is None or (type(key1) == bytes and len(key1) == 16)
+        assert isinstance(key2, np.datetime64)
+        assert isinstance(key3, uuid.UUID)
+
+        return key1 + dt_to_bytes(key2) + key3.bytes
+
+    def _deserialize_key(self, data):
+        assert data is None or (type(data) == bytes and len(data) == 40)
+
+        if data:
+            key1 = data[:16]
+            key2 = bytes_to_dt(data[16:24])
+            key3 = uuid.UUID(bytes=data[24:])
+        else:
+            key1 = b'\x00' * 16
+            key2 = np.datetime64(0, 'ns')
+            key3 = uuid.UUID(bytes=b'\x00' * 16)
+
+        return key1, key2, key3
+
+
 #
 # Value Types
 #
@@ -1312,3 +1425,42 @@ class _FlatBuffersValuesMixin(object):
 
     def _deserialize_value(self, data):
         return self._cast(data)
+
+
+class _Pickle5ValuesMixin(object):
+    """
+    Arbitrary Python object values, serialized using Pickle protocol version 5.
+    Protocol version 5 was added in Python 3.8. It adds support for out-of-band
+    data and speedup for in-band data.
+
+    .. seealso::
+
+        * https://docs.python.org/3/library/pickle.html#data-stream-format
+        * https://www.python.org/dev/peps/pep-0574/
+    """
+
+    PROTOCOL = 5
+
+    def _serialize_value(self, value):
+        obj_buffers = []
+        obj_data = pickle.dumps(value, protocol=self.PROTOCOL, buffer_callback=obj_buffers.append)
+        data = []
+        data.append(struct.pack('>I', len(obj_data)))
+        data.append(obj_data)
+        for d in obj_buffers:
+            data.append(struct.pack('>I', len(d.raw())))
+            data.append(d)
+        return b''.join(data)
+
+    def _deserialize_value(self, data):
+        data = memoryview(data)
+        obj_buffers = []
+        obj_len = struct.unpack('>I', data[0:4])[0]
+        obj_data = data[4:obj_len + 4]
+        i = obj_len + 4
+        while i < len(data):
+            buffer_len = struct.unpack('>I', data[i:i + 4])[0]
+            buffer_data = data[i + 4:i + 4 + buffer_len]
+            obj_buffers.append(buffer_data)
+            i += 4 + buffer_len
+        return pickle.loads(obj_data, buffers=obj_buffers)
